#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2018 Silvio Peroni (essepuntato@gmail.com)
# 
# Permission to use, copy, modify, and/or distribute this software for any purpose 
# with or without fee is hereby granted, provided that the above copyright notice 
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH 
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, 
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, 
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS 
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import argparse
import csv
import itertools
import operator
from scipy import stats
from numpy import mean, std
import collections
import sys
import os
from math import sqrt


grouping = {
    "self": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    "citation": [10, 20, 30, 40, 50, 60, 70, 80],
    "year": [1970, 1990, 2000, 2004, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016]
}


def group_by(csv_table, field):
    result = {}

    csv_list = sorted(csv_table, key=operator.itemgetter(field))
    for key, group in itertools.groupby(csv_list, lambda item: item[field]):
        if key not in result:
            result[key] = []
        result[key] += group

    return result


def __get_stats(csv_table, exclude_fields):
    result = {}

    for row in csv_table:
        for key in row:
            if key not in exclude_fields:
                if key not in result:
                    result[key] = {}
                if row[key] not in result[key]:
                    result[key][row[key]] = 0
                result[key][row[key]] += 1

    return result


def get_stats(csv_table, exclude_fields=[], group_by_field=None):
    if group_by_field is None:
        result = __get_stats(csv_table, exclude_fields)
    else:
        result = {}
        grouped = group_by(csv_table, group_by_field)
        for key in grouped:
            result[key] = __get_stats(grouped[key], exclude_fields)

    return result


def get_desc_stat(csv_table, field, value, min_v=None, max_v=None, less_than_or_equal_to=True,
                  exclude_fields=[], group_by_value={}, cast_function=lambda x: x, count_field=[]):
    result = ""
    stat = __get_stats(csv_table, exclude_fields)

    max_value = sys.maxsize
    min_value = -max_value - 1
    max_field_value = max_value if max_v is None else cast_function(max_v)
    min_field_value = min_value if min_v is None else cast_function(min_v)
    for key in stat:
        result += "\n%s" % key
        if key in count_field:
            total_value = 0
            for k, v in stat[key].items():
                total_value += cast_function(k) * cast_function(v)
            result += " (total: %s)" % total_value
        od = collections.OrderedDict(sorted(stat[key].items()))
        if key in group_by_value:
            thresholds = group_by_value[key]
            for idx, threshold in enumerate(thresholds):
                is_last = idx + 1 == len(thresholds)

                prev_threshold = min_value if idx == 0 else threshold if is_last else thresholds[idx - 1] + 1
                next_threshold = max_value if is_last else threshold

                if key != field or \
                        (less_than_or_equal_to and min_field_value <= prev_threshold <= cast_function(value)) or \
                        (not less_than_or_equal_to and cast_function(value) < next_threshold <= max_field_value):
                    cur_value = 0

                    for k, v in od.items():
                        if prev_threshold <= cast_function(k) <= next_threshold:
                            cur_value += cast_function(v)

                    if prev_threshold == next_threshold:
                        string_value = str(threshold)
                    elif prev_threshold == min_value:
                        string_value = "x <= " + str(threshold)
                    elif next_threshold == max_value:
                        string_value = "x >= " + str(threshold)
                    else:
                        string_value = str(prev_threshold) + " <= x <= " + str(next_threshold)
                    result += "\n\t%s: %s" % (string_value, cur_value)
        else:
            for k, v in od.items():
                result += "\n\t%s: %s" % (k, v)

    return result



def split_by(csv_table, field, value, min_value=None, max_value=None, cast_function=lambda x: x):
    less_or_equal_than = []
    greater_than = []
    cur_value = cast_function(value)
    cur_min_value = None if min_value is None else cast_function(min_value)
    cur_max_value = None if max_value is None else cast_function(max_value)

    for row in csv_table:
        row_value = cast_function(row[field])

        if (cur_min_value is None or row_value >= cur_min_value) and (cur_max_value is None or row_value <= cur_max_value):
            if row_value > cur_value:
                greater_than += [row]
            else:
                less_or_equal_than += [row]

    return less_or_equal_than, greater_than


def remove_duplicate_by_field(csv_table, field):
    result = []
    exising = set()

    for row in csv_table:
        id_value = row[field]
        if id_value not in exising:
            exising.add(id_value)
            result += [row]

    return result


def run_95_ci_statistics(csv_table):
    stats.t.interval(0.95, len(a) - 1, loc=mean(a), scale=stats.sem(a))


# http://hamelg.blogspot.com/2015/11/python-for-data-analysis-part-23-point.html
# https://scientificallysound.org/2017/05/16/independent-t-test-python/
def run_student_statistics(csv_table, id, field, value, min_value=None, max_value=None,
                           cast_function=lambda x: x, count_fields=[]):
    desc = ""

    # Create two groups of rows split by year value to 2012, and made unique according to the "id" field
    a, b = map(remove_duplicate_by_field,
               split_by(csv_table, field, value, min_value, max_value, cast_function), (id, id))

    stat = "\n\n## Descriptive statistics of population [%s <= %s]%s" \
           "\n\n## Descriptive statistics of population [%s > %s]%s" % \
           (field, value, get_desc_stat(list(a), field, value, min_value, max_value, True, ["id"], grouping,
                                        cast_function, count_fields),
            field, value, get_desc_stat(list(b), field, value, min_value, max_value, False, ["id"], grouping,
                                        cast_function, count_fields))

    # Consider only the values of the key "self" contained in the two groups
    v_a, v_b = map(lambda x: map(operator.itemgetter("self"), x), (list(a), list(b)))
    n_a = list(map(int, list(v_a)))
    mean_a = round(mean(n_a), 3)
    std_a = round(std(n_a), 3)
    if args.n:
        n_a = list(filter(lambda x: x != 0, n_a))
    desc += "Population [%s <= %s]: %s\n\tMean: %s\n\tStandard deviation: %s" % \
            (field, value, len(n_a), mean_a, std_a)

    n_b = list(map(int, list(v_b)))
    mean_b = round(mean(n_b), 3)
    std_b = round(std(n_b), 3)
    if args.n:
        n_b = list(filter(lambda x: x != 0, n_b))
    desc += "\nPopulation [%s > %s]: %s\n\tMean: %s\n\tStandard deviation: %s" % \
            (field, value, len(n_b), mean_b, std_b)

    diff_mean = round(mean_b - mean_a, 3)
    desc += "\nDifference of means ([%s > %s] - [%s <= %s]): %s" % \
            (field, value, field, value, diff_mean)

    # Calculate the mean difference and 95% confidence interval
    # (code modified from https://scientificallysound.org/2017/05/16/independent-t-test-python/)
    l_a = len(n_a)  # n of population A
    l_b = len(n_b)  # n of population B
    degree_of_freedom = (l_a + l_b - 2)
    average_std_ab = sqrt(((l_a - 1)*(std_a)**2 + (l_b - 1)*(std_b)**2) / degree_of_freedom)

    margin_of_error = stats.t.ppf(0.975, degree_of_freedom) * average_std_ab * sqrt(1/l_a + 1/l_b)
    ci_low = round(diff_mean - margin_of_error, 3)
    ci_high = round(diff_mean + margin_of_error, 3)

    desc += "\nConfidence interval (0.95): %s - %s%s\n---------------------------------------------------" % \
            (ci_low, ci_high, stat)

    return desc, {
        "# p[%s <= %s]" % (field, value): len(n_a),
        "mean p[%s <= %s]" % (field, value): mean_a,
        "st p[%s <= %s]" % (field, value): std_a,
        "# p[%s > %s]" % (field, value): len(n_b),
        "mean p[%s > %s]" % (field, value): mean_b,
        "st p[%s > %s]" % (field, value): std_b,
        "diff": diff_mean,
        "ci-low": ci_low,
        "ci-high": ci_high
    }


if __name__ == "__main__":

    arg_parser = argparse.ArgumentParser("analyse_data.py")
    arg_parser.add_argument("-i", "--input", dest="i", required=True,
                            help="The CSV file with all the citations and categories.")
    arg_parser.add_argument("-o", "--output", dest="o",
                            help="The file where to store the results.")
    arg_parser.add_argument("-f", "--id-field", dest="f", default="id",
                            help="The name of the field that identifies the citing resources.")
    arg_parser.add_argument("-y", "--year", dest="y", default="2012",
                            help="The year used to split the results in two gropus.")
    arg_parser.add_argument("-m", "--min", dest="m", default=None,
                            help="Minimum year to consider.")
    arg_parser.add_argument("-M", "--max", dest="M", default=None,
                            help="Maximum year to consider.")
    arg_parser.add_argument("-c", "--category", dest="c", default=None,
                            help="The category to analyse.")
    arg_parser.add_argument("-n", "--no-zero", dest="n", default=False, action="store_true",
                            help="Exclude the articles with 0 self-citations.")
    arg_parser.add_argument("-cc", "--counting-citation", dest="cc", default=["citation", "self"],
                            help="The fields for which a total amount should be returned in description.")

    args = arg_parser.parse_args()

    desc = ""

    with open(args.i) as f:
        csv_reader = list(csv.DictReader(f))
        # pprint.pprint(group_by(csv_reader, "category"))

        if args.c is None:
            csv_table = csv_reader
        else:
            csv_table = group_by(csv_reader, "category")[args.c]

        desc += "# Results (all)"
        print(desc)

        field = "year"
        value = args.y

        desc_sub, t_stat = run_student_statistics(csv_table, args.f, field, value, args.m, args.M, int, args.cc)
        desc += "\n" + desc_sub
        print(desc_sub)

        final_result = []

        if args.c is None:
            t_stat["category"] = "ALL"
            final_result += [t_stat]

            csv_table = group_by(csv_reader, "category")
            for category in csv_table:
                desc_c = "\n# Results for category '%s'" % category
                desc += "\n" + desc_c
                print(desc_c)

                desc_sub, t_stat = run_student_statistics(csv_table[category], args.f, field,
                                                          value, args.m, args.M, int, args.cc)
                desc += "\n" + desc_sub
                print(desc_sub)

                t_stat["category"] = category
                final_result += [t_stat]
        else:
            t_stat["category"] = args.c
            final_result += [t_stat]

    if args.o is not None and desc != "":
        base_dir = os.path.dirname(args.o)
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)

        with open(args.o, "w") as f:
            csv_writer = csv.DictWriter(f, ["category", "# p[%s <= %s]" % (field, value),
                                            "mean p[%s <= %s]" % (field, value),
                                            "st p[%s <= %s]" % (field, value), "# p[%s > %s]" % (field, value),
                                            "mean p[%s > %s]" % (field, value), "st p[%s > %s]" % (field, value),
                                            "diff", "ci-low", "ci-high"])
            csv_writer.writeheader()
            csv_writer.writerows(final_result)
        with open(args.o + ".desc.txt", "w") as f:
            f.write(desc)

