#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2018 Silvio Peroni (essepuntato@gmail.com)
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.linear_model import LinearRegression
import matplotlib.pyplot as plt
plt.close("all")
from os.path import exists
from os import makedirs
from scipy.stats import lognorm


def fit_string(s, limit):
    result = ""
    cur_char = 0

    for token in s.split():
        if cur_char + len(token) + 1 > limit:
            result += "\n"
            cur_char = 0
        result += " " + token
        cur_char += len(token) + 1

    return result.strip()


def df_per_category(df, cat, rng):
    df = df[df.category == cat]
    return df[df.year >= rng[0]][df.year <= rng[1]]


def figures_per_category(df, cat, rng):
    df_rng = df_per_category(df, cat, rng)
    df_rng_self = df_rng.self
    df_rng_self_mean = df_rng_self.mean()
    df_rng_self_count = df_rng_self.value_counts()
    return df_rng_self_count, df_rng_self_mean


def self_citations_plot(df, figure_title=""):
    fig, axes = plt.subplots(nrows=4, ncols=6)
    fig.suptitle(figure_title)
    plt.subplots_adjust(hspace=1.0, wspace=0.5)
    r, c = 0, 0
    min_year = df.year.unique().min()

    return fig, axes, r, c, min_year


def mean_figures_per_category(df, cat, rng, item_in_sample=20, n_samples=10000):
    result = []

    df_rng = df_per_category(df, cat, rng)
    for idx in range(n_samples):
        result.append(df_rng.sample(item_in_sample, replace=True).self.mean())

    result_series = pd.Series(result)
    return result_series.round().value_counts().sort_index(), result_series.mean(), result_series


def diff_mean_figures_per_category(df, cat, rng_old, rng_new, item_in_sample=20, n_samples=10000):
    result = []

    df_rng_new = df_per_category(df, cat, rng_new)
    df_rng_old = df_per_category(df, cat, rng_old)
    for idx in range(n_samples):
        sample_new = df_rng_new.sample(item_in_sample, replace=True).self.mean()
        sample_old = df_rng_old.sample(item_in_sample, replace=True).self.mean()
        result.append(sample_new - sample_old)

    result_series = pd.Series(result)
    return result_series.round().value_counts().sort_index(), result_series.mean(), result_series


def plot_self_citation_graph(fig, axes, r, c, category, vMi12, v0912, v1316, t):
    subf = fig.plot(kind=t, ax=axes[r, c], fontsize=8, color=["b", "g", "r"])
    subf.legend(loc=1, prop={'size': 5})
    subf.set_title(fit_string(category, 25), size="small", wrap=True)

    # Lines indicating the self-citation means for each period
    if vMi12:
        subf.axvline(vMi12[1], ls="dashed", lw=0.5, c="b")
    if v0912:
        subf.axvline(v0912[1], ls="dashed", lw=0.5, c="g")
    if v1316:
        subf.axvline(v1316[1], ls="dashed", lw=0.5, c="r")

    if c == 5:
        r, c = r + 1, 0
    else:
        c += 1

    return r, c


def self_citations_diff_means(df, figure_title):
    fig, axes, r, c, min_year = self_citations_plot(df, figure_title)

    base = "2018/JOI 2018/data/normal_diff_means/%s/" % figure_title
    if not exists(base):
        makedirs(base)

    for category in sorted(df.category.unique()):
        v0916 = diff_mean_figures_per_category(df, category, (2009, 2012), (2013, 2016))
        v0916[2].to_csv(path_or_buf="%s%s_v0916.csv" % (base, category), index=False)

        vMi16 = diff_mean_figures_per_category(df, category, (min_year, 2012), (2013, 2016))
        vMi16[2].to_csv(path_or_buf="%s%s_vMi16.csv" % (base, category), index=False)

        # DataFrame with the self-citations of the category
        sdf = pd.DataFrame([vMi16[0], v0916[0]]).transpose()
        sdf.columns = [str(min_year)[2:] + "-12\\13-16", "09-12\\13-16"]
        r, c = plot_self_citation_graph(sdf, axes, r, c, category, v0916, vMi16, None, "line")
        print(category, r, c)


def self_citations_means(df, figure_title):
    fig, axes, r, c, min_year = self_citations_plot(df, figure_title)

    base = "2018/JOI 2018/data/lognormal_means/%s/" % figure_title
    if not exists(base):
        makedirs(base)

    for category in sorted(df.category.unique()):
        vMi12 = mean_figures_per_category(df, category, (min_year, 2012))
        vMi12[2].to_csv(path_or_buf="%s%s_vMi12.csv" % (base, category), index=False)

        v0912 = mean_figures_per_category(df, category, (2009, 2012))
        v0912[2].to_csv(path_or_buf="%s%s_v0912.csv" % (base, category), index=False)

        v1316 = mean_figures_per_category(df, category, (2013, 2016))
        v1316[2].to_csv(path_or_buf="%s%s_v1316.csv" % (base, category), index=False)

        # DataFrame with the self-citations of the category
        sdf = pd.DataFrame([vMi12[0], v0912[0], v1316[0]]).transpose()
        sdf.columns = [str(min_year)[2:] + "-12", "09-12", "13-16"]
        r, c = plot_self_citation_graph(sdf, axes, r, c, category, vMi12, v0912, v1316, "line")
        print(category, r, c)


def self_citations_graph(df, figure_title, n_x_items=10):
    fig, axes, r, c, min_year = self_citations_plot(df, figure_title)

    for category in sorted(df.category.unique()):
        # This draws the number of articles with 0 or more self-citations according to
        # each discipline considered in the input dataset
        vMi12 = figures_per_category(df, category, (min_year, 2012))
        v0912 = figures_per_category(df, category, (2009, 2012))
        v1316 = figures_per_category(df, category, (2013, 2016))

        # DataFrame with the self-citations of the category
        sdf = pd.DataFrame(
            [vMi12[0][:n_x_items], v0912[0][:n_x_items], v1316[0][:n_x_items]]).transpose()
        sdf.columns = [str(min_year)[2:] + "-12", "09-12", "13-16"]
        r, c = plot_self_citation_graph(sdf, axes, r, c, category, vMi12, v0912, v1316, "bar")


def category_confidence_interval(category, rng, figure_title):
    # Calculate the 95% confidence interval (CI) of the distribution of
    # the category for the range of years to consider
    file_path = "2018/JOI 2018/data/lognormal_means/%s/%s_v%s%s.csv" % \
                (figure_title, category, "Mi" if rng[0] < 2000 else str(rng[0])[2:], str(rng[1])[2:])
    means = pd.read_csv(file_path)
    shape, location, scale = lognorm.fit(means.values)
    mean = lognorm.mean(shape, loc=location, scale=scale)
    ci_min, ci_max = lognorm.interval(0.95, shape, loc=location, scale=scale)

    return mean, ci_min, ci_max


def linear_regression(df, rng, rng_new):
    min_rng, max_rng = rng

    # The mean of the self-citations grouped by year
    y = df[min_rng <= df.year][df.year <= max_rng].groupby(by="year").mean().sort_index()

    # The year in consideration to use to calculate the linear regression
    x = np.array(y.index.unique().sort_values()).reshape(-1, 1)

    # The predictive model for the value, based on the aforementioned x and y
    model = LinearRegression().fit(x, y.self.values)

    # The years that we want to compare with the linear regression line
    min_rng_new, max_rng_new = rng_new
    new_x = np.array(range(min_rng_new, max_rng_new + 1)).reshape(-1, 1)
    # The prediction of the means for each year of the aforementioned years if
    pred_y = model.predict(new_x)

    return new_x, pred_y


def plot_linear_regression_with_ci_per_category(category, df, rng, rng_new, ax):
    cat_df = df[df.category == category]
    x, pred_y = linear_regression(cat_df, rng, rng_new)
    x, pred_y2 = linear_regression(cat_df, rng_new, rng_new)

    scatter_df = cat_df[rng_new[0] <= cat_df.year][cat_df.year <= rng_new[1]].\
        groupby(by="year").mean().reset_index(level=0)

    # means per year of the new range of dates
    subf = scatter_df.plot(x="year", y="self", xticks=range(rng_new[0], rng_new[1] + 1),
                           kind="scatter", ax=ax, fontsize=8)
    subf.set_xlabel("")
    subf.set_ylabel("")
    subf.set_title(fit_string(category, 25), size="small", wrap=True)

    # linear regression of the previous range of dates
    subf.plot(x, pred_y, color='red')

    # linear regression of the new range of dates
    subf.plot(x, pred_y2, color='blue')


def plot_linear_regression_with_ci(df, rng, rng_new, figure_title):
    fig, axes, r, c, min_year = self_citations_plot(df, figure_title)

    for category in sorted(df.category.unique()):
        plot_linear_regression_with_ci_per_category(category, df, rng, rng_new, axes[r, c])

        if c == 5:
            r, c = r + 1, 0
        else:
            c += 1


def plot_error_diagram(df, figure_title):
    fig, axes, r, c, min_year = self_citations_plot(df, figure_title)

    for category in sorted(df.category.unique()):
        ci = category_confidence_interval(category, (2013, 2016), figure_title)

        if c == 5:
            r, c = r + 1, 0
        else:
            c += 1



# csv_file = "../data/author-self-citations.csv"
author_self_citation_file = "/Users/six/Dropbox/Università/PhD/Papers/2018/JOI 2018/data/author-self-citations.csv"
author_network_self_citation_file = "/Users/six/Dropbox/Università/PhD/Papers/2018/JOI 2018/data/author-network-self-citations.csv"

# self_citations_graph(pd.read_csv(author_self_citation_file), "Author self-citations", 9)
# self_citations_graph(pd.read_csv(author_network_self_citation_file),
#                     "Author network self-citations", 6)

# self_citations_means(pd.read_csv(author_self_citation_file), "Author self-citations")
# self_citations_means(pd.read_csv(author_network_self_citation_file), "Author network self-citations")

# self_citations_diff_means(pd.read_csv(author_self_citation_file), "Author self-citations")
# self_citations_diff_means(pd.read_csv(author_network_self_citation_file), "Author network self-citations")

# plot_linear_regression_with_ci(
#     pd.read_csv(author_self_citation_file), (2009, 2012), (2013, 2016), "Author self-citations")
# plot_linear_regression_with_ci(
#     pd.read_csv(author_self_citation_file), (1957, 2012), (2013, 2016), "Author self-citations")

# plot_linear_regression_with_ci(
#     pd.read_csv(author_network_self_citation_file), (2009, 2012), (2013, 2016),
#     "Author network self-citations")
# plot_linear_regression_with_ci(
#     pd.read_csv(author_network_self_citation_file), (1957, 2012), (2013, 2016),
#     "Author network self-citations")


# x = np.array(range(2009, 2017)).reshape(-1, 1)
# y = df[df.pre == "-"][df.year >= 2009].groupby(by="year").mean().self.values
# model = LinearRegression().fit(x[:4], y)
# y_pred = model.predict(x)
# y2 = df[df.pre == "+"].groupby(by="year").mean().self.values
# plt.scatter(x[4:], y2)
# plt.plot(x, y_pred, color='red')
# plt.scatter(x[:4], y, color='red')
#
# shape, location, scale = lognorm.fit(rm)  # rm = the list of all the means, with repetition or pure means in case of values
# mu, sigma = np.log(scale), shape
# lognorm.interval(0.95, shape, loc=location, scale=scale)
# lognorm.mean(shape, loc=location, scale=scale)
